@@ -24,9 +24,7 @@ Boston, MA 02111-1307, USA.  */
 #ifdef WINDOWSNT
 /* See comments in lastfile.c.  */
 char my_begdata[] = "Beginning of Emacs initialized data";
-
-/* The MSVC linker seems to segregate static vs public globals.  */
-char my_begbss[1];
+char my_begbss[1];  /* Do not initialize this variable.  */
 static char _my_begbss[1];
 char * my_begbss_static = _my_begbss;
 
