@@ -55,9 +55,7 @@ Boston, MA 02111-1307, USA.
 #undef read
 #undef write
 
-#define getwd _getwd
 #include "lisp.h"
-#undef getwd
 
 #include <pwd.h>
 
@@ -79,20 +77,29 @@ Boston, MA 02111-1307, USA.
 #include "nt.h"
 #include "ndir.h"
 #include "ntheap.h"
-  
+
 
 extern Lisp_Object Vwin32_downcase_file_names;
 extern Lisp_Object Vwin32_generate_fake_inodes;
 extern Lisp_Object Vwin32_get_true_file_attributes;
 
+static char startup_dir[ MAXPATHLEN ];
 
 /* Get the current working directory.  */
 char *
 getwd (char *dir)
 {
+#if 0
   if (GetCurrentDirectory (MAXPATHLEN, dir) > 0)
     return dir;
   return NULL;
+#else
+  /* Emacs doesn't actually change directory itself, and we want to
+     force our real wd to be where emacs.exe is to avoid unnecessary
+     conflicts when trying to rename or delete directories.  */
+  strcpy (dir, startup_dir);
+  return dir;
+#endif
 }
 
 #ifndef HAVE_SOCKETS
@@ -379,9 +386,10 @@ unixtodos_filename (p)
    (From msdos.c...probably should figure out a way to share it,
    although this code isn't going to ever change.)  */
 int
-crlf_to_lf (n, buf)
+crlf_to_lf (n, buf, lf_count)
      register int n;
      register unsigned char *buf;
+     register unsigned *lf_count;
 {
   unsigned char *np = buf;
   unsigned char *startp = buf;
@@ -391,6 +399,8 @@ crlf_to_lf (n, buf)
     return n;
   while (buf < endp - 1)
     {
+      if (*buf == 0x0a)
+	(*lf_count)++;
       if (*buf == 0x0d)
 	{
 	  if (*(++buf) != 0x0a)
@@ -400,7 +410,11 @@ crlf_to_lf (n, buf)
 	*np++ = *buf++;
     }
   if (buf < endp)
-    *np++ = *buf++;
+    {
+      if (*buf == 0x0a)
+	(*lf_count)++;
+      *np++ = *buf++;
+    }
   return np - startp;
 }
 
@@ -450,6 +464,23 @@ get_long_basename (char * name, char * buf, int size)
   WIN32_FIND_DATA find_data;
   HANDLE dir_handle;
   int len = 0;
+#ifdef PIGSFLY
+  char *p;
+
+  /* If the last component of NAME has a wildcard character, 
+     return it as the basename.  */
+  p = name + strlen (name);
+  while (*p != '\\' && *p != ':' && p > name) p--;
+  if (p > name) p++;
+  if (strchr (p, '*') || strchr (p, '?'))
+    {
+      if ((len = strlen (p)) < size)
+	memcpy (buf, p, len + 1);
+      else
+	len = 0;
+      return len;
+    }
+#endif
 
   dir_handle = FindFirstFile (name, &find_data);
   if (dir_handle != INVALID_HANDLE_VALUE)
@@ -624,7 +655,9 @@ init_environment ()
       "EMACSDATA",
       "EMACSPATH",
       "EMACSLOCKDIR",
-      "INFOPATH",
+      /* We no longer set INFOPATH because Info-default-directory-list
+	 is then ignored.  We use a hook in winnt.el instead.  */
+      /*      "INFOPATH", */
       "EMACSDOC",
       "TERM",
     };
@@ -674,6 +707,26 @@ init_environment ()
 	memcpy (*envp, "COMSPEC=", 8);
   }
 
+  /* Remember the initial working directory for getwd, then make the
+     real wd be the location of emacs.exe to avoid conflicts when
+     renaming or deleting directories.  (We also don't call chdir when
+     running subprocesses for the same reason.)  */
+  if (!GetCurrentDirectory (MAXPATHLEN, startup_dir))
+    abort ();
+
+  {
+    char *p;
+    char modname[MAX_PATH];
+
+    if (!GetModuleFileName (NULL, modname, MAX_PATH))
+      abort ();
+    if ((p = strrchr (modname, '\\')) == NULL)
+      abort ();
+    *p = 0;
+
+    SetCurrentDirectory (modname);
+  }
+
   init_user_info ();
 }
 
@@ -1000,6 +1053,7 @@ map_win32_filename (const char * name, const char ** pPath)
   char * str = shortname;
   char c;
   char * path;
+  const char * save_name = name;
 
   if (is_fat_volume (name, &path)) /* truncate to 8.3 */
     {
@@ -1084,7 +1138,7 @@ map_win32_filename (const char * name, const char ** pPath)
     }
 
   if (pPath)
-    *pPath = shortname + (path - name);
+    *pPath = shortname + (path - save_name);
 
   return shortname;
 }
@@ -2400,6 +2454,7 @@ sys_read (int fd, char * buffer, unsigned int count)
 	  *buffer++ = 0x0d;
 	  count--;
 	  nchars++;
+	  fd_info[fd].flags &= ~FILE_LAST_CR;
 	}
 
       /* presence of a child_process structure means we are operating in
@@ -2415,8 +2470,10 @@ sys_read (int fd, char * buffer, unsigned int count)
 	    {
 	    case STATUS_READ_FAILED:
 	    case STATUS_READ_ERROR:
-	      /* report normal EOF */
-	      return 0;
+	      /* report normal EOF if nothing in buffer */
+	      if (nchars <= 0)
+		fd_info[fd].flags |= FILE_AT_EOF;
+	      return nchars;
 
 	    case STATUS_READ_READY:
 	    case STATUS_READ_IN_PROGRESS:
@@ -2445,8 +2502,9 @@ sys_read (int fd, char * buffer, unsigned int count)
 	    {
 	      PeekNamedPipe ((HANDLE) _get_osfhandle (fd), NULL, 0, NULL, &waiting, NULL);
 	      to_read = min (waiting, (DWORD) count);
-      
-	      nchars += _read (fd, buffer, to_read);
+
+	      if (to_read > 0)
+		nchars += _read (fd, buffer, to_read);
 	    }
 #ifdef HAVE_SOCKETS
 	  else /* FILE_SOCKET */
@@ -2478,12 +2536,21 @@ sys_read (int fd, char * buffer, unsigned int count)
 #endif
 	}
       else
-	nchars += _read (fd, buffer, count);
-
-      /* Perform text mode translation if required.  */
-      if ((fd_info[fd].flags & FILE_BINARY) == 0)
 	{
-	  nchars = crlf_to_lf (nchars, orig_buffer);
+	  int nread = _read (fd, buffer, count);
+	  if (nread >= 0)
+	    nchars += nread;
+	  else if (nchars == 0)
+	    nchars = nread;
+	}
+
+      if (nchars <= 0)
+	fd_info[fd].flags |= FILE_AT_EOF;
+      /* Perform text mode translation if required.  */
+      else if ((fd_info[fd].flags & FILE_BINARY) == 0)
+	{
+	  unsigned lf_count = 0;
+	  nchars = crlf_to_lf (nchars, orig_buffer, &lf_count);
 	  /* If buffer contains only CR, return that.  To be absolutely
 	     sure we should attempt to read the next char, but in
 	     practice a CR to be followed by LF would not appear by
@@ -2493,8 +2560,6 @@ sys_read (int fd, char * buffer, unsigned int count)
 	      fd_info[fd].flags |= FILE_LAST_CR;
 	      nchars--;
 	    }
-	  else
-	    fd_info[fd].flags &= ~FILE_LAST_CR;
 	}
     }
   else
