@@ -50,6 +50,10 @@ extern int change_frame_size (FRAME_PTR, int, int, int, int);
 /* from w32fns.c */
 extern Lisp_Object Vwin32_alt_is_meta;
 
+/* from w32term.c */
+extern Lisp_Object Vwin32_capslock_is_shiftlock;
+extern Lisp_Object Vwin32_recognize_altgr;
+
 /* Event queue */
 #define EVENT_QUEUE_SIZE 50
 static INPUT_RECORD event_queue[EVENT_QUEUE_SIZE];
@@ -96,12 +100,14 @@ get_frame (void)
 /* Translate console modifiers to emacs modifiers.  
    German keyboard support (Kai Morgan Zeise 2/18/95).  */
 int
-win32_kbd_mods_to_emacs (DWORD mods)
+win32_kbd_mods_to_emacs (DWORD mods, WORD key)
 {
   int retval = 0;
 
-  /* If AltGr has been pressed, remove it.  */
-  if ((mods & (RIGHT_ALT_PRESSED | LEFT_CTRL_PRESSED)) 
+  /* If we recognize right-alt and left-ctrl as AltGr, and it has been
+     pressed, remove the modifiers.  */
+  if (!NILP (Vwin32_recognize_altgr) 
+      && (mods & (RIGHT_ALT_PRESSED | LEFT_CTRL_PRESSED)) 
       == (RIGHT_ALT_PRESSED | LEFT_CTRL_PRESSED))
     mods &= ~ (RIGHT_ALT_PRESSED | LEFT_CTRL_PRESSED);
 
@@ -116,9 +122,36 @@ win32_kbd_mods_to_emacs (DWORD mods)
 	retval |= meta_modifier;
     }
 
-  if (((mods & (SHIFT_PRESSED | CAPSLOCK_ON)) == SHIFT_PRESSED)
-      || ((mods & (SHIFT_PRESSED | CAPSLOCK_ON)) == CAPSLOCK_ON))
-    retval |= shift_modifier;
+  /* Just in case someone wanted the original behaviour, make it
+     optional by setting win32-capslock-is-shiftlock to t.  */
+  if (NILP (Vwin32_capslock_is_shiftlock) &&
+#if 1
+      ( (key == VK_BACK) ||
+	(key == VK_TAB) ||
+	(key == VK_CLEAR) ||
+	(key == VK_RETURN) ||
+	(key == VK_ESCAPE) ||
+	( (key >= VK_SPACE) && (key <= VK_HELP)) ||
+	( (key >= VK_NUMPAD0) && (key <= VK_F24))
+	)
+#else
+      /* Perhaps easier to say which keys we *do* always want affected
+	 by capslock.  Not sure how this affects "alphabetic" keyboard
+	 input in non-English languages though - what virtual key codes
+	 are returned for accented letters, for instance?  */
+      !( (key >= '0' && key <= '9') || (key >= 'A' && key <= 'Z') )
+#endif
+      )
+    {
+      if ( (mods & SHIFT_PRESSED) == SHIFT_PRESSED)
+	retval |= shift_modifier;
+    }
+  else
+    {
+      if (((mods & (SHIFT_PRESSED | CAPSLOCK_ON)) == SHIFT_PRESSED)
+	  || ((mods & (SHIFT_PRESSED | CAPSLOCK_ON)) == CAPSLOCK_ON))
+	retval |= shift_modifier;
+    }
 
   return retval;
 }
@@ -147,7 +180,10 @@ win32_kbd_patch_key (KEY_EVENT_RECORD *event)
     keystate[VK_SHIFT] = 0x80;
   if (mods & CAPSLOCK_ON) 
     keystate[VK_CAPITAL] = 1;
-  if ((mods & LEFT_CTRL_PRESSED) && (mods & RIGHT_ALT_PRESSED))
+  /* If we recognize right-alt and left-ctrl as AltGr, set the key
+     states accordingly before invoking ToAscii.  */
+  if (!NILP (Vwin32_recognize_altgr)
+      && (mods & LEFT_CTRL_PRESSED) && (mods & RIGHT_ALT_PRESSED))
     {
       keystate[VK_CONTROL] = 0x80;
       keystate[VK_LCONTROL] = 0x80;
@@ -379,7 +415,8 @@ key_event (KEY_EVENT_RECORD *event, struct input_event *emacs_ev, int *isdead)
 #else
   XSETFRAME (emacs_ev->frame_or_window, get_frame ());
 #endif
-  emacs_ev->modifiers = win32_kbd_mods_to_emacs (event->dwControlKeyState);
+  emacs_ev->modifiers = win32_kbd_mods_to_emacs (event->dwControlKeyState,
+                                                 event->wVirtualKeyCode);
   emacs_ev->timestamp = GetTickCount ();
   if (key_flag == 2) return -1; /* 95.7.25 by himi */
   return 1;
@@ -486,7 +523,7 @@ do_mouse_event (MOUSE_EVENT_RECORD *event,
   
   button_state = event->dwButtonState;
   emacs_ev->timestamp = GetTickCount ();
-  emacs_ev->modifiers = win32_kbd_mods_to_emacs (event->dwControlKeyState) |
+  emacs_ev->modifiers = win32_kbd_mods_to_emacs (event->dwControlKeyState, 0) |
     ((event->dwButtonState & mask) ? down_modifier : up_modifier);
   
   XSETFASTINT (emacs_ev->x, event->dwMousePosition.X);
