@@ -45,6 +45,7 @@ Boston, MA 02111-1307, USA.
 #include "systime.h"
 #include "syswait.h"
 #include "process.h"
+#include "w32term.h"
 
 /* Control whether spawnve quotes arguments as necessary to ensure
    correct parsing by child process.  Because not all uses of spawnve
@@ -54,7 +55,13 @@ Lisp_Object Vwin32_quote_process_args;
 
 /* Control whether create_child causes the process' window to be
    hidden.  The default is nil. */
-Lisp_Object Vw32_start_process_show_window;
+Lisp_Object Vwin32_start_process_show_window;
+
+/* Control whether create_child causes the process to inherit Emacs'
+   console window, or be given a new one of its own.  The default is
+   nil, to allow multiple DOS programs to run on Win95.  Having separate
+   consoles also allows Emacs to cleanly terminate process groups.  */
+Lisp_Object Vwin32_start_process_share_console;
 
 /* Time to sleep before reading from a subprocess output pipe - this
    avoids the inefficiency of frequently reading small amounts of data.
@@ -77,6 +84,8 @@ Lisp_Object Vwin32_generate_fake_inodes;
    are supported on NTFS volumes, this is only relevant on NT.  */
 Lisp_Object Vwin32_get_true_file_attributes;
 
+Lisp_Object Qhigh, Qlow;
+
 #ifndef SYS_SIGLIST_DECLARED
 extern char *sys_siglist[];
 #endif
@@ -283,6 +292,11 @@ reader_thread (void *arg)
   return 0;
 }
 
+/* To avoid Emacs changing directory, we just record here the directory
+   the new process should start in.  This is set just before calling
+   sys_spawnve, and is not generally valid at any other time.  */
+static char * process_dir;
+
 static BOOL 
 create_child (char *exe, char *cmdline, char *env,
 	      int * pPid, child_process *cp)
@@ -290,6 +304,7 @@ create_child (char *exe, char *cmdline, char *env,
   STARTUPINFO start;
   SECURITY_ATTRIBUTES sec_attrs;
   SECURITY_DESCRIPTOR sec_desc;
+  char dir[ MAXPATHLEN ];
   
   if (cp == NULL) abort ();
   
@@ -297,7 +312,7 @@ create_child (char *exe, char *cmdline, char *env,
   start.cb = sizeof (start);
   
 #ifdef HAVE_NTGUI
-  if (NILP (Vw32_start_process_show_window))
+  if (NILP (Vwin32_start_process_show_window))
     start.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
   else
     start.dwFlags = STARTF_USESTDHANDLES;
@@ -316,14 +331,15 @@ create_child (char *exe, char *cmdline, char *env,
   sec_attrs.nLength = sizeof (sec_attrs);
   sec_attrs.lpSecurityDescriptor = &sec_desc;
   sec_attrs.bInheritHandle = FALSE;
+
+  strcpy (dir, process_dir);
+  unixtodos_filename (dir);
   
   if (!CreateProcess (exe, cmdline, &sec_attrs, NULL, TRUE,
-#ifdef SHARE_CONSOLE
-		      CREATE_NEW_PROCESS_GROUP,
-#else
-		      CREATE_NEW_CONSOLE,
-#endif
-		      env, NULL,
+		      (!NILP (Vwin32_start_process_share_console)
+		       ? CREATE_NEW_PROCESS_GROUP
+		       : CREATE_NEW_CONSOLE),
+		      env, dir,
 		      &start, &cp->procinfo))
     goto EH_Fail;
 
@@ -513,7 +529,6 @@ get_result:
     {
       *status = retval;
     }
-#if 0
   else if (synch_process_alive)
     {
       synch_process_alive = 0;
@@ -539,7 +554,6 @@ get_result:
 
       reap_subprocess (cp);
     }
-#endif
 
   reap_subprocess (cp);
   
@@ -593,8 +607,13 @@ win32_executable_type (char * filename, int * is_dos_app, int * is_cygnus_app)
 
       nt_header = (PIMAGE_NT_HEADERS) ((char *) dos_header + dos_header->e_lfanew);
 
-      if (nt_header->Signature != IMAGE_NT_SIGNATURE &&
-	  LOWORD (nt_header->Signature) != IMAGE_OS2_SIGNATURE)
+      if ((char *) nt_header > (char *) dos_header + executable.size) 
+	{
+	  /* Some dos headers (pkunzip) have bogus e_lfanew fields.  */
+	  *is_dos_app = TRUE;
+	} 
+      else if (nt_header->Signature != IMAGE_NT_SIGNATURE &&
+		 LOWORD (nt_header->Signature) != IMAGE_OS2_SIGNATURE)
 	{
 	  *is_dos_app = TRUE;
 	}
@@ -1091,8 +1110,17 @@ sys_select (int nfds, SELECT_TYPE *rfds, SELECT_TYPE *wfds, SELECT_TYPE *efds,
 	    else
 	      {
 		/* Unable to find something to wait on for this fd, skip */
+
+		/* Note that this is not a fatal error, and can in fact
+		   happen in unusual circumstances.  Specifically, if
+		   sys_spawnve fails, eg. because the program doesn't
+		   exist, and debug-on-error is t so Fsignal invokes a
+		   nested input loop, then the process output pipe is
+		   still included in input_wait_mask with no child_proc
+		   associated with it.  (It is removed when the debugger
+		   exits the nested input loop and the error is thrown.)  */
+
 		DebPrint (("sys_select: fd %ld is invalid! ignoring\n", i));
-		abort ();
 	      }
 	  }
       }
@@ -1101,8 +1129,14 @@ count_children:
   /* Add handles of child processes.  */
   nc = 0;
   for (cp = child_procs+(child_proc_count-1); cp >= child_procs; cp--)
-    /* some child_procs might be sockets; ignore them */
-    if (CHILD_ACTIVE (cp) && cp->procinfo.hProcess)
+    /* Some child_procs might be sockets; ignore them.  Also some
+       children may have died already, but we haven't finished reading
+       the process output; ignore them too.  */
+    if (CHILD_ACTIVE (cp) && cp->procinfo.hProcess
+	&& (cp->fd < 0
+	    || (fd_info[cp->fd].flags & FILE_SEND_SIGCHLD) == 0
+	    || (fd_info[cp->fd].flags & FILE_AT_EOF) != 0)
+	)
       {
 	wait_hnd[nh + nc] = cp->procinfo.hProcess;
 	cps[nc] = cp;
@@ -1158,9 +1192,16 @@ count_children:
       if (active >= nh)
 	{
 	  cp = cps[active - nh];
+
+	  /* We cannot always signal SIGCHLD immediately; if we have not
+	     finished reading the process output, we must delay sending
+	     SIGCHLD until we do.  */
+
+	  if (cp->fd >= 0 && (fd_info[cp->fd].flags & FILE_AT_EOF) == 0)
+	    fd_info[cp->fd].flags |= FILE_SEND_SIGCHLD;
 	  /* SIG_DFL for SIGCHLD is ignore */
-	  if (sig_handlers[SIGCHLD] != SIG_DFL &&
-	      sig_handlers[SIGCHLD] != SIG_IGN)
+	  else if (sig_handlers[SIGCHLD] != SIG_DFL &&
+		   sig_handlers[SIGCHLD] != SIG_IGN)
 	    {
 #ifdef FULL_DEBUG
 	      DebPrint (("select calling SIGCHLD handler for pid %d\n",
@@ -1211,8 +1252,6 @@ count_children:
 
 /* Substitute for certain kill () operations */
 
-#ifndef SHARE_CONSOLE
-
 static BOOL CALLBACK
 find_child_console (HWND hwnd, child_process * cp)
 {
@@ -1238,8 +1277,6 @@ find_child_console (HWND hwnd, child_process * cp)
   return TRUE;
 }
 
-#endif /* !SHARE_CONSOLE */
-
 int 
 sys_kill (int pid, int sig)
 {
@@ -1271,25 +1308,13 @@ sys_kill (int pid, int sig)
       proc_hand = cp->procinfo.hProcess;
       pid = cp->procinfo.dwProcessId;
 
-#ifndef SHARE_CONSOLE
       /* Try to locate console window for process. */
       EnumWindows (find_child_console, (LPARAM) cp);
-#endif
     }
   
   if (sig == SIGINT)
     {
-#ifdef SHARE_CONSOLE
-      /* Ctrl-Break is NT equivalent of SIGINT.  */
-      if (!GenerateConsoleCtrlEvent (CTRL_BREAK_EVENT, pid))
-        {
-	  DebPrint (("sys_kill.GenerateConsoleCtrlEvent return %d "
-		     "for pid %lu\n", GetLastError (), pid));
-	  errno = EINVAL;
-	  rc = -1;
-        }
-#else
-      if (cp && cp->hwnd)
+      if (NILP (Vwin32_start_process_share_console) && cp && cp->hwnd)
 	{
 	  BYTE control_scan_code = (BYTE) MapVirtualKey (VK_CONTROL, 0);
 	  BYTE vk_break_code = VK_CANCEL;
@@ -1312,15 +1337,25 @@ sys_kill (int pid, int sig)
 	      keybd_event (vk_break_code, break_scan_code, KEYEVENTF_KEYUP, 0);
 	      keybd_event (VK_CONTROL, control_scan_code, KEYEVENTF_KEYUP, 0);
 
+	      /* Sleep for a bit to give time for Emacs frame to respond
+		 to focus change events (if Emacs was active app).  */
+	      Sleep (10);
+
 	      SetForegroundWindow (foreground_window);
 	    }
 	}
-#endif
+      /* Ctrl-Break is NT equivalent of SIGINT.  */
+      else if (!GenerateConsoleCtrlEvent (CTRL_BREAK_EVENT, pid))
+        {
+	  DebPrint (("sys_kill.GenerateConsoleCtrlEvent return %d "
+		     "for pid %lu\n", GetLastError (), pid));
+	  errno = EINVAL;
+	  rc = -1;
+        }
     }
   else
     {
-#ifndef SHARE_CONSOLE
-      if (cp && cp->hwnd)
+      if (NILP (Vwin32_start_process_share_console) && cp && cp->hwnd)
 	{
 #if 1
 	  if (os_subtype == OS_WIN95)
@@ -1341,18 +1376,33 @@ sys_kill (int pid, int sig)
    Could try to invoke DestroyVM through CallVxD.
 
 */
+#if 0
+	      /* On Win95, posting WM_QUIT causes the 16-bit subsystem
+		 to hang when cmdproxy is used in conjunction with
+		 command.com for an interactive shell.  Posting
+		 WM_CLOSE pops up a dialog that, when Yes is selected,
+		 does the same thing.  TerminateProcess is also less
+		 than ideal in that subprocesses tend to stick around
+		 until the machine is shutdown, but at least it
+		 doesn't freeze the 16-bit subsystem.  */
 	      PostMessage (cp->hwnd, WM_QUIT, 0xff, 0);
+#endif
+	      if (!TerminateProcess (proc_hand, 0xff))
+		{
+		  DebPrint (("sys_kill.TerminateProcess returned %d "
+			     "for pid %lu\n", GetLastError (), pid));
+		  errno = EINVAL;
+		  rc = -1;
+		}
 	    }
 	  else
 #endif
 	    PostMessage (cp->hwnd, WM_CLOSE, 0, 0);
 	}
-      else
-#endif
       /* Kill the process.  On Win32 this doesn't kill child processes
 	 so it doesn't work very well for shells which is why it's not
 	 used in every case.  */
-      if (!TerminateProcess (proc_hand, 0xff))
+      else if (!TerminateProcess (proc_hand, 0xff))
         {
 	  DebPrint (("sys_kill.TerminateProcess returned %d "
 		     "for pid %lu\n", GetLastError (), pid));
@@ -1451,6 +1501,12 @@ reset_standard_handles (int in, int out, int err, HANDLE handles[3])
   SetStdHandle (STD_ERROR_HANDLE, handles[2]);
 }
 
+void
+set_process_dir (char * dir)
+{
+  process_dir = dir;
+}
+
 #ifdef HAVE_SOCKETS
 
 /* To avoid problems with winsock implementations that work over dial-up
@@ -1520,7 +1576,8 @@ socket connections still exist.")
 #endif /* HAVE_SOCKETS */
 
 
-/* Some miscellaneous functions that are Window specific.  */
+/* Some miscellaneous functions that are Windows specific, but not GUI
+   specific (ie. are applicable in terminal or batch mode as well).  */
 
 /* lifted from fileio.c  */
 #define CORRECT_DIR_SEPS(s) \
@@ -1575,15 +1632,212 @@ All path elements in FILENAME are converted to their long names.")
   return build_string (longname);
 }
 
+DEFUN ("win32-set-process-priority", Fwin32_set_process_priority, Swin32_set_process_priority,
+       2, 2, 0,
+  "Set the priority of PROCESS to PRIORITY.\n\
+If PROCESS is nil, the priority of Emacs is changed, otherwise the\n\
+priority of the process whose pid is PROCESS is changed.\n\
+PRIORITY should be one of the symbols high, normal, or low;\n\
+any other symbol will be interpreted as normal.\n\
+\n\
+If successful, the return value is t, otherwise nil.")
+  (process, priority)
+     Lisp_Object process, priority;
+{
+  HANDLE proc_handle = GetCurrentProcess ();
+  DWORD  priority_class = NORMAL_PRIORITY_CLASS;
+  Lisp_Object result = Qnil;
+
+  CHECK_SYMBOL (priority, 0);
+
+  if (!NILP (process))
+    {
+      DWORD pid;
+      child_process *cp;
+
+      CHECK_NUMBER (process, 0);
+
+      /* Allow pid to be an internally generated one, or one obtained
+	 externally.  This is necessary because real pids on Win95 are
+	 negative.  */
+
+      pid = XINT (process);
+      cp = find_child_pid (pid);
+      if (cp != NULL)
+	pid = cp->procinfo.dwProcessId;
+
+      proc_handle = OpenProcess (PROCESS_SET_INFORMATION, FALSE, pid);
+    }
+
+  if (EQ (priority, Qhigh))
+    priority_class = HIGH_PRIORITY_CLASS;
+  else if (EQ (priority, Qlow))
+    priority_class = IDLE_PRIORITY_CLASS;
+
+  if (proc_handle != NULL)
+    {
+      if (SetPriorityClass (proc_handle, priority_class))
+	result = Qt;
+      if (!NILP (process))
+	CloseHandle (proc_handle);
+    }
+
+  return result;
+}
+
+
+DEFUN ("win32-get-locale-info", Fwin32_get_locale_info, Swin32_get_locale_info, 1, 2, 0,
+  "Return information about the Windows locale LCID.\n\
+By default, return a three letter locale code which encodes the default\n\
+language as the first two characters, and the country or regionial variant\n\
+as the third letter.  For example, ENU refers to `English (United States)',\n\
+while ENC means `English (Canadian)'.\n\
+\n\
+If the optional argument LONGFORM is non-nil, the long form of the locale\n\
+name is returned, e.g. `English (United States)' instead.\n\
+\n\
+If LCID (a 16-bit number) is not a valid locale, the result is nil.")
+     (lcid, longform)
+     Lisp_Object lcid, longform;
+{
+  int got_abbrev;
+  int got_full;
+  char abbrev_name[32] = { 0 };
+  char full_name[256] = { 0 };
+
+  CHECK_NUMBER (lcid, 0);
+
+  if (!IsValidLocale (XINT (lcid), LCID_SUPPORTED))
+    return Qnil;
+
+  if (NILP (longform))
+    {
+      got_abbrev = GetLocaleInfo (XINT (lcid),
+				  LOCALE_SABBREVLANGNAME | LOCALE_USE_CP_ACP,
+				  abbrev_name, sizeof (abbrev_name));
+      if (got_abbrev)
+	return build_string (abbrev_name);
+    }
+  else
+    {
+      got_full = GetLocaleInfo (XINT (lcid),
+				LOCALE_SLANGUAGE | LOCALE_USE_CP_ACP,
+				full_name, sizeof (full_name));
+      if (got_full)
+	return build_string (full_name);
+    }
+
+  return Qnil;
+}
+
+
+DEFUN ("win32-get-current-locale-id", Fwin32_get_current_locale_id, Swin32_get_current_locale_id, 0, 0, 0,
+  "Return Windows locale id for current locale setting.\n\
+This is a numerical value; use `win32-get-locale-info' to convert to a\n\
+human-readable form.")
+     ()
+{
+  return make_number (GetThreadLocale ());
+}
+
+
+DEFUN ("win32-get-default-locale-id", Fwin32_get_default_locale_id, Swin32_get_default_locale_id, 0, 1, 0,
+  "Return Windows locale id for default locale setting.\n\
+By default, the system default locale setting is returned; if the optional\n\
+parameter USERP is non-nil, the user default locale setting is returned.\n\
+This is a numerical value; use `win32-get-locale-info' to convert to a\n\
+human-readable form.")
+     (userp)
+     Lisp_Object userp;
+{
+  if (NILP (userp))
+    return make_number (GetSystemDefaultLCID ());
+  return make_number (GetUserDefaultLCID ());
+}
+
+DWORD int_from_hex (char * s)
+{
+  DWORD val = 0;
+  static char hex[] = "0123456789abcdefABCDEF";
+  char * p;
+
+  while (*s && (p = strchr(hex, *s)) != NULL)
+    {
+      unsigned digit = p - hex;
+      if (digit > 15)
+	digit -= 6;
+      val = val * 16 + digit;
+      s++;
+    }
+  return val;
+}
+
+/* We need to build a global list, since the EnumSystemLocale callback
+   function isn't given a context pointer.  */
+Lisp_Object Vwin32_valid_locale_ids;
+
+BOOL CALLBACK enum_locale_fn (LPTSTR localeNum)
+{
+  DWORD id = int_from_hex (localeNum);
+  Vwin32_valid_locale_ids = Fcons (make_number (id), Vwin32_valid_locale_ids);
+  return TRUE;
+}
+
+DEFUN ("win32-get-valid-locale-ids", Fwin32_get_valid_locale_ids, Swin32_get_valid_locale_ids, 0, 0, 0,
+  "Return list of all valid Windows locale ids.\n\
+Each id is a numerical value; use `win32-get-locale-info' to convert to a\n\
+human-readable form.")
+     ()
+{
+  Vwin32_valid_locale_ids = Qnil;
+
+  EnumSystemLocales (enum_locale_fn, LCID_SUPPORTED);
+
+  Vwin32_valid_locale_ids = Fnreverse (Vwin32_valid_locale_ids);
+  return Vwin32_valid_locale_ids;
+}
+
+
+DEFUN ("win32-set-current-locale", Fwin32_set_current_locale, Swin32_set_current_locale, 1, 1, 0,
+  "Make Windows locale LCID be the current locale setting for Emacs.\n\
+If successful, the new locale id is returned, otherwise nil.")
+     (lcid)
+     Lisp_Object lcid;
+{
+  CHECK_NUMBER (lcid, 0);
+
+  if (!IsValidLocale (XINT (lcid), LCID_SUPPORTED))
+    return Qnil;
+
+  if (!SetThreadLocale (XINT (lcid)))
+    return Qnil;
+
+  /* Need to set input thread locale if present.  */
+  if (dwWinThreadId)
+    /* Reply is not needed.  */
+    PostThreadMessage (dwWinThreadId, WM_EMACS_SETLOCALE, XINT (lcid), 0);
+
+  return make_number (GetThreadLocale ());
+}
+
 
 syms_of_ntproc ()
 {
+  Qhigh = intern ("high");
+  Qlow = intern ("low");
+
 #ifdef HAVE_SOCKETS
   defsubr (&Swin32_has_winsock);
   defsubr (&Swin32_unload_winsock);
 #endif
   defsubr (&Swin32_short_file_name);
   defsubr (&Swin32_long_file_name);
+  defsubr (&Swin32_set_process_priority);
+  defsubr (&Swin32_get_locale_info);
+  defsubr (&Swin32_get_current_locale_id);
+  defsubr (&Swin32_get_default_locale_id);
+  defsubr (&Swin32_get_valid_locale_ids);
+  defsubr (&Swin32_set_current_locale);
 
   DEFVAR_LISP ("win32-quote-process-args", &Vwin32_quote_process_args,
     "Non-nil enables quoting of process arguments to ensure correct parsing.\n\
@@ -1597,11 +1851,21 @@ quote characters that appear, otherwise a suitable escape character\n\
 will be chosen based on the type of the program.");
   Vwin32_quote_process_args = Qt;
 
-  DEFVAR_LISP ("w32-start-process-show-window",
-	       &Vw32_start_process_show_window,
+  DEFVAR_LISP ("win32-start-process-show-window",
+	       &Vwin32_start_process_show_window,
     "When nil, processes started via start-process hide their windows.\n\
 When non-nil, they show their window in the method of their choice.");
-  Vw32_start_process_show_window = Qnil;
+  Vwin32_start_process_show_window = Qnil;
+
+  DEFVAR_LISP ("win32-start-process-share-console",
+	       &Vwin32_start_process_share_console,
+    "When nil, processes started via start-process are given a new console.\n\
+When non-nil, they share the Emacs console; this has the limitation of\n\
+allowing only only DOS subprocess to run at a time (whether started directly\n\
+or indirectly by Emacs), and preventing Emacs from cleanly terminating the\n\
+subprocess group, but may allow Emacs to interrupt a subprocess that doesn't\n\
+otherwise respond to interrupts from Emacs.");
+  Vwin32_start_process_share_console = Qnil;
 
   DEFVAR_INT ("win32-pipe-read-delay", &Vwin32_pipe_read_delay,
     "Forced delay before reading subprocess output.\n\
