@@ -331,6 +331,7 @@ close_file_data (file_data *p_file)
 
 /* Routines to manipulate NT executable file sections.  */
 
+#ifdef SEPARATE_BSS_SECTION
 static void
 get_bss_info_from_map_file (file_data *p_infile, PUCHAR *p_bss_start, 
 			    DWORD *p_bss_size)
@@ -369,6 +370,7 @@ get_bss_info_from_map_file (file_data *p_infile, PUCHAR *p_bss_start,
   *p_bss_start = (PUCHAR) start;
   *p_bss_size = (DWORD) len;
 }
+#endif
 
 unsigned long
 get_section_size (PIMAGE_SECTION_HEADER p_section)
@@ -484,7 +486,7 @@ get_section_info (file_data *p_infile)
 	  data_size = my_edata - data_start_va;
 	}
 #else
-      if (!strcmp (section->Name, "APPDATA")) 
+      if (!strcmp (section->Name, "EMDATA")) 
 	{
 	  /* The Emacs initialized data section.  */
 	  data_section = section;
@@ -516,6 +518,18 @@ get_section_info (file_data *p_infile)
 	+ data_section->VirtualAddress;
     }
 #else
+/* As noted in lastfile.c, the Alpha (but not the Intel) MSVC linker
+   globally segregates all static and public bss data (ie. across all
+   linked modules, not just per module), so we must take both static and
+   public bss areas into account to determine the true extent of the bss
+   area used by Emacs.
+
+   To be strictly correct, we should dump the static and public bss
+   areas used by Emacs separately if non-overlapping (since otherwise we
+   are dumping bss data belonging to system libraries, eg. the static
+   bss system data on the Alpha).  However, in practice this doesn't
+   seem to matter, since presumably the system libraries always
+   reinitialize their bss variables.  */
   bss_start = min (my_begbss, my_begbss_static);
   bss_size = max (my_endbss, my_endbss_static) - bss_start;
 #endif
