@@ -66,6 +66,8 @@ extern Lisp_Object Qoverriding_local_map, Qoverriding_terminal_local_map;
 
 extern Lisp_Object Qmenu_bar_update_hook;
 
+void set_frame_menubar ();
+
 static Lisp_Object win32_dialog_show ();
 static Lisp_Object win32menu_show ();
 
@@ -1225,7 +1227,8 @@ cached information about equivalent key sequences.")
   if (! NILP (position))
     {
       /* Decode the first argument: find the window and the coordinates.  */
-      if (EQ (position, Qt))
+      if (EQ (position, Qt)
+          || (CONSP (position) && EQ (XCONS (position)->car, Qmenu_bar)))
 	{
 	  /* Use the mouse's current position.  */
 	  FRAME_PTR new_f = 0;
@@ -1483,11 +1486,36 @@ get_frame_menubar_event (f, num)
   
   return (Qnil);
 }
+
+/* Activate the menu bar of frame F.
+   This is called from keyboard.c when it gets the
+   menu_bar_activate_event out of the Emacs event queue.
+
+   To activate the menu bar, we signal to the input thread that it can
+   return from the WM_INITMENU message, allowing the normal Windows
+   processing of the menus.
+
+   But first we recompute the menu bar contents (the whole tree).
+
+   This way we can safely execute Lisp code.  */
+   
+x_activate_menubar (f)
+     FRAME_PTR f;
+{
+  set_frame_menubar (f, 0, 1);
+
+  /* Lock out further menubar changes while active.  */
+  f->output_data.win32->menubar_active = 1;
+
+  /* Signal input thread to return from WM_INITMENU.  */
+  complete_deferred_msg (FRAME_WIN32_WINDOW (f), WM_INITMENU, 0);
+}
 
 void 
-set_frame_menubar (f, first_time)
+set_frame_menubar (f, first_time, deep_p)
      FRAME_PTR f;
      int first_time;
+     int deep_p;
 {
   Lisp_Object tail, items;
   HMENU hmenu;
@@ -1499,6 +1527,16 @@ set_frame_menubar (f, first_time)
   struct buffer *prev = current_buffer;
   Lisp_Object buffer;
 
+  /* We must not change the menubar when actually in use.  */
+  if (f->output_data.win32->menubar_active)
+    return;
+
+#if 0  /* I don't see why this should be needed */
+  /* Ensure menubar is up to date when about to be used.  */
+  if (f->output_data.win32->pending_menu_activation && !deep_p)
+    deep_p = 1;
+#endif
+
   buffer = XWINDOW (FRAME_SELECTED_WINDOW (f))->buffer;
   specbind (Qinhibit_quit, Qt);
   /* Don't let the debugger step into this code
@@ -1530,9 +1568,18 @@ set_frame_menubar (f, first_time)
   if (NILP (items))
     items = FRAME_MENU_BAR_ITEMS (f) = menu_bar_items (FRAME_MENU_BAR_ITEMS (f));
   
-  hmenu = CreateMenu ();
-  
-  if (!hmenu) goto error;
+  hmenu = f->output_data.win32->menubar_widget;
+  if (!hmenu)
+    {
+    hmenu = CreateMenu ();
+    if (!hmenu) goto error;
+    }
+  else
+    {
+      /* Delete current contents.  */
+      while (DeleteMenu (hmenu, 0, MF_BYPOSITION))
+	;
+    }
   
   discard_menu_items (&mm);
   UNBLOCK_INPUT;
@@ -1589,15 +1636,16 @@ set_frame_menubar (f, first_time)
   
   BLOCK_INPUT;
   {
-    HMENU old = GetMenu (FRAME_WIN32_WINDOW (f));
+    HMENU old = f->output_data.win32->menubar_widget;
     SetMenu (FRAME_WIN32_WINDOW (f), hmenu);
-    DestroyMenu (old);
+    f->output_data.win32->menubar_widget = hmenu;
+    /* Causes flicker when menu bar is updated 
+    DrawMenuBar (FRAME_WIN32_WINDOW (f)); */
 
     /* Force the window size to be recomputed so that the frame's text
        area remains the same, if menubar has just been created.  */
     if (old == NULL)
-      
-x_set_window_size (f, 0, FRAME_WIDTH (f), FRAME_HEIGHT (f));
+      x_set_window_size (f, 0, FRAME_WIDTH (f), FRAME_HEIGHT (f));
   }
   
  error:
@@ -1616,6 +1664,7 @@ free_frame_menubar (f)
   {
     HMENU old = GetMenu (FRAME_WIN32_WINDOW (f));
     SetMenu (FRAME_WIN32_WINDOW (f), NULL);
+    f->output_data.win32->menubar_widget = NULL;
     DestroyMenu (old);
   }
     
@@ -1629,7 +1678,7 @@ void
 initialize_frame_menubar (f)
      FRAME_PTR f;
 {
-  set_frame_menubar (f, 1);
+  set_frame_menubar (f, 1, 1);
 }
 
 #if 0
