@@ -213,6 +213,9 @@ extern struct win32_display_info *win32_term_init ();
 
 struct win32_output
 {
+  /* Menubar "widget" handle.  */
+  HMENU menubar_widget;
+
   /* Original palette (used to deselect real palette after drawing) */
   HPALETTE old_palette;
 
@@ -312,6 +315,13 @@ struct win32_output
 
   /* Nonzero means tried already to make this frame visible.  */
   char asked_for_visible;
+
+  /* Nonzero means menubar is currently active.  */
+  char menubar_active;
+
+  /* Nonzero means menubar is about to become active, but should be
+     brought up to date first.  */
+  volatile char pending_menu_activation;
 };
 
 /* Get at the computed faces of an X window frame.  */
@@ -591,7 +601,10 @@ extern void win32_unload_font ();
 #define WM_EMACS_SETWINDOWPOS          (WM_EMACS_START + 0x05)
 #define WM_EMACS_DESTROYWINDOW         (WM_EMACS_START + 0x06)
 #define WM_EMACS_TRACKPOPUPMENU        (WM_EMACS_START + 0x07)
-#define WM_EMACS_END                   (WM_EMACS_START + 0x10)
+#define WM_EMACS_SETFOCUS              (WM_EMACS_START + 0x08)
+#define WM_EMACS_SETFOREGROUND         (WM_EMACS_START + 0x09)
+#define WM_EMACS_SETLOCALE             (WM_EMACS_START + 0x0a)
+#define WM_EMACS_END                   (WM_EMACS_START + 0x0b)
 
 #define WND_FONTWIDTH_INDEX    (0) 
 #define WND_LINEHEIGHT_INDEX   (4) 
@@ -613,6 +626,17 @@ typedef struct Win32Msg {
     RECT rect;
 } Win32Msg;
 
+/* Structure for recording message when input thread must return a
+   result that depends on lisp thread to compute.  Lisp thread can
+   complete deferred messages out of order.  */
+typedef struct deferred_msg
+{
+  struct deferred_msg * next;
+  Win32Msg              w32msg;
+  LRESULT               result;
+  int                   completed;
+} deferred_msg;
+
 extern CRITICAL_SECTION critsect;
 
 extern void init_crit ();
@@ -628,6 +652,7 @@ extern int release_frame_dc (struct frame * f, HDC hDC);
 
 extern BOOL get_next_msg ();
 extern BOOL post_msg ();
+extern void complete_deferred_msg (HWND hwnd, UINT msg, LRESULT result);
 extern void wait_for_sync ();
 
 extern BOOL parse_button ();
